import json

from django.http import HttpResponse
from django.utils.translation import gettext_lazy as _
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View

from .wrappers import PaypalWrapper
from .models import PaypalPayment
from django_paypal import settings as django_paypal_settings


class NotifyPaypalView(View):
    paypal_wrapper = PaypalWrapper(auth={
        'API_CLIENT_ID': django_paypal_settings.PAYPAL_API_CLIENT_ID,
        'API_SECRET': django_paypal_settings.PAYPAL_API_SECRET,
    })

    def post(self, request, *args, **kwargs):
        request_data = json.loads(request.body.decode('utf-8'))

        # return HttpResponse(status=200)
        # general attributes
        if 'resource' not in request_data:
            return HttpResponse(status=400)
        payment_id = None
        for key in ['parent_payment', 'id']:
            if key in request_data['resource']:
                payment_id = request_data['resource'][key]
                break
        if payment_id is None:
            return HttpResponse(status=400)

        try:
            paypal_payment = PaypalPayment.objects.get(payment_id=payment_id)
        except PaypalPayment.DoesNotExist:
            return HttpResponse(status=400)
        return self.handle_updated_payment(paypal_payment=paypal_payment, expected_status='approved')

    @csrf_exempt
    def dispatch(self, request, *args, **kwargs):
        return super(NotifyPaypalView, self).dispatch(request, *args, **kwargs)

    def handle_updated_payment(self, paypal_payment, expected_status=None):
        """
            Override to use the paypal_payment in the way you want.
        """
        updated_payment = paypal_payment
        if updated_payment.refresh_from_paypal(self.paypal_wrapper, expected_status=expected_status):
            if updated_payment.status == 'approved':
                import logging
                logger = logging.getLogger(__name__)
                logger.error(_('Paypal: Status of checkout {} is now {}').format(updated_payment.payment_id, updated_payment.status))
                return HttpResponse(status=400)
            return HttpResponse(status=200)
        return HttpResponse(status=400)



from __future__ import unicode_literals

import json
import logging

from decimal import Decimal
from django.db import models
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import gettext_lazy as _


@python_2_unicode_compatible
class PaypalPayment(models.Model):
    payment_id = models.CharField(_("payment id"), max_length=255, unique=True)  # id created by paypal
    intent = models.CharField(_("intent"), max_length=255)  # sale
    experience_profile_id = models.CharField(_("intent"), max_length=255, blank=True)  # sale
    note_to_payer = models.CharField(_("intent"), max_length=165, blank=True)  # sale
    state = models.CharField(_("state"), max_length=255, blank=True)
    payment_method = models.CharField(_("payment method"), max_length=255, blank=True)
    custom = models.CharField(_("payment method"), max_length=255, blank=True)
    payer_id = models.CharField(_("payer id"), max_length=255, blank=True)
    transaction_fee = models.DecimalField(_("transaction fee"), max_digits=9, decimal_places=2, default=Decimal(0))
    related_resource_id = models.CharField(_("related resource id"), max_length=255, blank=True, null=True)
    initial_response_object = models.TextField(_("initial post response"), null=True, blank=True)
    update_response_object = models.TextField(_("updated get response"), null=True, blank=True)

    failure_reason = models.CharField(_("failure reason"), max_length=255, blank=True)

    link = models.URLField(_("link"), blank=True)
    approve_link = models.URLField(_("approve link"), blank=True)
    execute_link = models.URLField(_("execute link"), blank=True)

    created_at = models.DateTimeField(_("created at"), auto_now_add=True)
    last_modified = models.DateTimeField(_("last modified"), auto_now=True)

    objects = models.Manager()

    def __str__(self):
        return self.payment_id

    class Meta:
        verbose_name = _("Paypal Payment")
        verbose_name_plural = _("Paypal Payments")

    def execute(self, paypal_wrapper):
        if not self.execute_link:
            return False
        if self.state == 'approved':
            return False
        if not self.payer_id:
            return False
        execute_response = paypal_wrapper.call_api(url=self.execute_link, data={'payer_id': self.payer_id})
        if execute_response and 'state' in execute_response and execute_response['state'] == 'approved':
            self.state = 'approved'
            self.save()
            return True
        return False

    def refresh_from_paypal(self, paypal_wrapper, expected_status=None):
        payment_response = paypal_wrapper.call_api(url=self.link)
        if not payment_response:
            logger = logging.getLogger(__name__)
            logger.error("Paypal Payment Link not available: {}".format(self.link))
            return False

        if expected_status and expected_status != payment_response['state']:
            logger = logging.getLogger(__name__)
            logger.error("Paypal Payment Status Error: expected: {0}, found: {1}".format(expected_status, payment_response['state']))
            return False

        self.update_response_object = json.dumps(payment_response)

        if payment_response['state'] == 'approved':
            fee = Decimal(0)
            for transaction in payment_response['transactions']:
                if 'related_resources' in transaction:
                    for resource in transaction['related_resources']:
                        if 'sale' in resource:
                            if 'links' in resource['sale']:
                                for link in resource['sale']['links']:
                                    if 'rel' in link and link['rel'] == 'self':
                                        sale_response = paypal_wrapper.call_api(url=link['href'])
                                        if sale_response and 'transaction_fee' in sale_response and 'value' in sale_response['transaction_fee'] and sale_response['transaction_fee']['currency'] == 'EUR':
                                            self.transaction_fee = Decimal(sale_response['transaction_fee']['value'])
                            if 'id' in resource['sale']:
                                self.related_resource_id = resource['sale']['id']

        if 'payer' in payment_response and 'payer_info' in payment_response['payer'] and 'payer_id' in payment_response['payer']['payer_info']:
            self.payer_id = payment_response['payer']['payer_info']['payer_id']

        if 'failure_reason' in payment_response:
            self.failure_reason = payment_response['failure_reason']

        self.state = payment_response['state']

        if 'links' in payment_response:
            for link in payment_response['links']:
                if 'rel' in link:
                    if link['rel'] == "self":
                        self.link = link['href']
                    elif link['rel'] == "approval_url":
                        self.approve_link = link['href']
                    elif link['rel'] == "execute":
                        self.execute_link = link['href']
        self.save()

        return self


@python_2_unicode_compatible
class PaypalTransaction(models.Model):
    payment = models.ForeignKey(PaypalPayment, verbose_name=_("payment"), related_name='transactions', on_delete=models.PROTECT)

    total_amount = models.DecimalField(_("total amount"), max_digits=9, decimal_places=2)
    currency = models.CharField(_("currency"), max_length=10)
    subtotal = models.DecimalField(_("subtotal"), max_digits=9, decimal_places=2)
    tax = models.DecimalField(_("tax"), max_digits=9, decimal_places=2, blank=True)
    shipping = models.DecimalField(_("shipping"), max_digits=9, decimal_places=2, blank=True)
    handling_fee = models.DecimalField(_("handling fee"), max_digits=9, decimal_places=2, blank=True)
    shipping_discount = models.DecimalField(_("shipping discount"), max_digits=9, decimal_places=2, blank=True)
    insurance = models.DecimalField(_("insurance fee"), max_digits=9, decimal_places=2, blank=True)
    gift_wrap = models.DecimalField(_("gift wrap fee"), max_digits=9, decimal_places=2, blank=True)

    reference_id = models.CharField(_("reference id"), max_length=255, blank=True)
    settlement_destination = models.CharField(_("settlement destination"), max_length=255, default="PARTNER_BALANCE")
    allowed_payment_method = models.CharField(_("allowed payment method"), max_length=255, default="INSTANT_FUNDING_SOURCE")  # UNRESTRICTED or INSTANT_FUNDING_SOURCE or IMMEDIATE_PAY
    description = models.CharField(_("description"), max_length=255, blank=True)
    note_to_payee = models.CharField(_("note to payee"), max_length=255, blank=True)
    custom = models.CharField(_("custom"), max_length=127, blank=True)
    invoice_number = models.CharField(_("invoice number"), max_length=127)
    purchase_order = models.CharField(_("purchase order"), max_length=127)
    soft_descriptor = models.CharField(_("soft descriptor"), max_length=22)

    created_at = models.DateTimeField(_("created at"), auto_now_add=True)
    last_modified = models.DateTimeField(_("last modified"), auto_now=True)

    shipping_method = models.CharField(_("shipping method"), max_length=255, blank=True)
    shipping_phone_number = models.CharField(_("shipping phone number"), max_length=255, blank=True)

    shipping_address_line_1 = models.CharField(_("shipping address line 1"), max_length=100)  # number, street, etc.
    shipping_address_line_2 = models.CharField(_("shipping address line 2"), max_length=100, blank=True)  # apt number, etc.
    shipping_address_city = models.CharField(_("shipping address city"), max_length=50)
    shipping_address_country = models.CharField(_("shipping address country"), max_length=2)
    shipping_address_postal_code = models.CharField(_("shipping address postal code"), max_length=20, blank=True)  # required in certain countries
    shipping_address_state = models.CharField(_("shipping address state"), max_length=100, blank=True)
    shipping_address_phone = models.CharField(_("shipping address phone"), max_length=50)
    shipping_address_normalization_status = models.CharField(_("shipping address normalization status"), max_length=50, default="UNKNOWN")  # UNKNOWN or UNNORMALIZED_USER_PREFERRED or NORMALIZED or UNNORMALIZED
    shipping_address_type = models.CharField(_("shipping address type"), max_length=50, default="HOME_OR_WORK")  # HOME_OR_WORK or GIFT or other
    shipping_address_recipient_name = models.CharField(_("shipping address recipient name"), max_length=127)

    objects = models.Manager()

    def __str__(self):
        return '{} - {} {}'.format(self.reference_id, self.total_amount, self.currency)

    class Meta:
        verbose_name = _("Paypal Transaction")
        verbose_name_plural = _("Paypal Transactions")


@python_2_unicode_compatible
class PaypalItem(models.Model):
    transaction = models.ForeignKey(PaypalTransaction, verbose_name=_("transaction"), related_name='items', on_delete=models.PROTECT)

    sku = models.CharField(_("stock keeping unit"), max_length=127)
    name = models.CharField(_("name"), max_length=127)
    description = models.CharField(_("description"), max_length=127, blank=True)
    quantity = models.CharField(_("quantity"), max_length=10)
    price = models.CharField(_("price"), max_length=10)
    currency = models.CharField(_("currency"), max_length=3)
    tax = models.CharField(_("currency"), max_length=10, blank=True)
    url = models.URLField(_("url"), blank=True)

    created_at = models.DateTimeField(_("created at"), auto_now_add=True)
    last_modified = models.DateTimeField(_("last modified"), auto_now=True)

    objects = models.Manager()

    def __str__(self):
        return self.name

    class Meta:
        verbose_name = _("Paypal Item")
        verbose_name_plural = _("Paypal Item")



from __future__ import unicode_literals

import base64
import json
import logging
import os
import sys
from decimal import Decimal

import certifi
from django.conf import settings

from django_paypal import settings as django_paypal_settings
from django_paypal.models import PaypalTransaction, PaypalPayment, PaypalItem
from django_paypal.utils import build_paypal_full_uri

try:
    # For Python 3.0 and later
    from urllib.error import HTTPError
    from urllib.request import urlopen
    from urllib.request import Request
except ImportError:
    # Fall back to Python 2's urllib2
    from urllib2 import HTTPError, Request, urlopen


class PaypalWrapper(object):
    interface_version = 'django_paypal_v{}'.format(django_paypal_settings.DJANGO_PAYPAL_VERSION)

    api_url = django_paypal_settings.PAYPAL_API_URL
    sandbox_url = django_paypal_settings.PAYPAL_SANDBOX_API_URL
    payment_url = django_paypal_settings.PAYPAL_PAYMENT_URL
    auth_url = django_paypal_settings.PAYPAL_AUTH_URL

    auth = None

    def __init__(self, auth=None, sandbox=None):
        super(PaypalWrapper, self).__init__()
        if getattr(settings, 'PAYPAL', False):
            self.auth = auth
            if sandbox is not None:
                if sandbox:
                    self.api_url = self.sandbox_url
            else:
                if django_paypal_settings.PAYPAL_SANDBOX:
                    self.api_url = self.sandbox_url

    def init(self, intent, payer, transactions, note_to_payer=None, experience_profile_id=None,
             success_url=django_paypal_settings.PAYPAL_SUCCESS_URL,
             cancellation_url=django_paypal_settings.PAYPAL_CANCELLATION_URL,
             notification_url=django_paypal_settings.PAYPAL_NOTIFICATION_URL):

        if not self.auth:
            return False

        # for transaction in transactions:
        #     transaction['notify_url'] = build_paypal_full_uri(notification_url)
        payment_data = {
            'intent': intent,
            'payer': payer,
            'transactions': transactions
        }
        if experience_profile_id:
            payment_data.update({'experience_profile_id': experience_profile_id})
        if note_to_payer:
            payment_data.update({'note_to_payer': note_to_payer})
        if note_to_payer:
            payment_data.update(
                {'redirect_urls':
                    {'return_url': build_paypal_full_uri(success_url),
                     'cancel_url': build_paypal_full_uri(cancellation_url)}
                 })

        payment_response = self.call_api(url=self.payment_url, data=payment_data)

        if payment_response and 'state' in payment_response and payment_response['state'] == 'created':

            transaction_array = []
            items_dict = {}
            try:
                paypal_payment = PaypalPayment(
                    state=payment_response['state'],
                    payment_id=payment_response['id'],
                    intent=payment_response['intent'],
                    experience_profile_id=payment_response.get('experience_profile_id', ''),
                    note_to_payer=payment_response.get('note_to_payer', ''),
                    payment_method=payment_response['payer'].get('payment_method', ''),
                    custom=payment_response.get('custom', ''),
                    initial_response_object=json.dumps(payment_response),
                )
                if 'links' in payment_response:
                    for link in payment_response['links']:
                        if 'rel' in link:
                            if link['rel'] == "self":
                                paypal_payment.link = link['href']
                            elif link['rel'] == "approval_url":
                                paypal_payment.approve_link = link['href']
                            elif link['rel'] == "execute":
                                paypal_payment.execute_link = link['href']

                for transaction in payment_response['transactions']:
                    transaction_array.append(PaypalTransaction(
                        reference_id=transaction['reference_id'],
                        total_amount=Decimal(transaction['amount']['total']),
                        currency=transaction['amount']['currency'],
                        subtotal=Decimal(transaction['amount']['details'].get('subtotal', 0)),
                        tax=Decimal(transaction['amount']['details'].get('tax', 0)),
                        shipping=Decimal(transaction['amount']['details'].get('shipping', 0)),
                        handling_fee=Decimal(transaction['amount']['details'].get('handling_fee', 0)),
                        shipping_discount=Decimal(transaction['amount']['details'].get('shipping_discount', 0)),
                        insurance=Decimal(transaction['amount']['details'].get('insurance', 0)),
                        gift_wrap=Decimal(transaction['amount']['details'].get('gift_wrap', 0)),
                        settlement_destination=transaction.get('settlement_destination', ''),
                        allowed_payment_method=transaction['payment_options'].get('allowed_payment_method', ''),
                        description=transaction.get('description', ''),
                        note_to_payee=transaction.get('note_to_payee', ''),
                        custom=transaction.get('custom', ''),
                        invoice_number=transaction.get('invoice_number', ''),
                        purchase_order=transaction.get('purchase_order', ''),
                        soft_descriptor=transaction.get('soft_descriptor', ''),
                        shipping_method=transaction['item_list'].get('shipping_method', ''),
                        shipping_phone_number=transaction['item_list'].get('shipping_phone_number', ''),
                        shipping_address_line_1=transaction['item_list']['shipping_address'].get('line_1', ''),
                        shipping_address_line_2=transaction['item_list']['shipping_address'].get('line_2', ''),
                        shipping_address_city=transaction['item_list']['shipping_address'].get('city', ''),
                        shipping_address_country=transaction['item_list']['shipping_address'].get('country', ''),
                        shipping_address_postal_code=transaction['item_list']['shipping_address'].get('postal_code', ''),
                        shipping_address_state=transaction['item_list']['shipping_address'].get('state', ''),
                        shipping_address_phone=transaction['item_list']['shipping_address'].get('phone', ''),
                        shipping_address_normalization_status=transaction['item_list']['shipping_address'].get('normalization_status', ''),
                        shipping_address_type=transaction['item_list']['shipping_address'].get('type', ''),
                        shipping_address_recipient_name=transaction['item_list']['shipping_address'].get('recipient_name', ''),
                    ))
                    for item in transaction['item_list']['items']:
                        if transaction['reference_id'] not in items_dict:
                            items_dict[transaction['reference_id']] = []
                        items_dict[transaction['reference_id']].append(PaypalItem(
                            sku=item.get('sku', ''),
                            name=item.get('name', ''),
                            description=item['description'],
                            quantity=item['quantity'],
                            price=item['price'],
                            currency=item['currency'],
                            tax=item.get('tax', ''),
                            url=item.get('url', '')
                        ))
            except KeyError as e:
                return False
            paypal_payment.save()
            for transaction in transaction_array:
                transaction.payment = paypal_payment
                transaction.save()
                if transaction.reference_id in items_dict:
                    for item in items_dict[transaction.reference_id]:
                        item.transaction = transaction
                        item.save()
            return paypal_payment
        else:
            return False

    def call_api(self, url=None, access_token=None, data=None):
        if not self.auth:
            return False
        if not url.lower().startswith('http'):
            url = '{0}{1}'.format(self.api_url, url)
        request = Request(url)

        if access_token is None:
            access_token = self._get_access_token()
        request.add_header('Authorization', 'Bearer {0}'.format(access_token))
        if data:
            data = json.dumps(data)
            data_len = len(data)
            request.add_header('Content-Length', data_len)
            request.data = data.encode(encoding='utf-8')
        elif data == '':
            request.method = 'POST'
            request.data = ''.encode(encoding='utf-8')
        request.add_header('Content-Type', 'application/json')
        try:
            if sys.version_info.major > 2 or (sys.version_info.major == 2 and sys.version_info.major > 7 or (sys.version_info.major == 7 and sys.version_info.major >= 9)):
                response = urlopen(request, cafile=certifi.where())
            else:
                response = urlopen(request)
        except HTTPError as e:
            logger = logging.getLogger(__name__)
            fp = e.fp
            body = fp.read()
            fp.close()
            if hasattr(e, 'code'):
                logger.error("Paypal Error {0}({1}): {2}".format(e.code, e.msg, body))
            else:
                logger.error("Paypal Error({0}): {1}".format(e.msg, body))
        else:
            return json.loads(response.read().decode('utf-8'))
        return False

    def _get_access_token(self):
        url = '{0}{1}'.format(self.api_url, self.auth_url)
        request = Request(url=url)
        # preparing request
        base64string = base64.encodestring(('%s:%s' % (self.auth['API_CLIENT_ID'], self.auth['API_SECRET'])).encode()).decode().replace('\n', '')
        request.add_header("Authorization", "Basic %s" % base64string)
        data = "grant_type=client_credentials"
        data_len = len(data)
        request.add_header('Accept', 'application/json')
        request.add_header('Content-Length', data_len)
        request.data = data.encode(encoding='utf-8')

        try:
            if sys.version_info.major > 2 or (sys.version_info.major == 2 and sys.version_info.major > 7 or (sys.version_info.major == 7 and sys.version_info.major >= 9)):
                response = urlopen(request, cafile=certifi.where())
            else:
                response = urlopen(request)
        except HTTPError as e:
            logger = logging.getLogger(__name__)
            fp = e.fp
            body = fp.read()
            fp.close()
            if hasattr(e, 'code'):
                logger.error("Paypal Error {0}({1}): {2}".format(e.code, e.msg, body))
            else:
                logger.error("Paypal Error({0}): {1}".format(e.msg, body))
        else:
            return json.loads(response.read().decode('utf-8'))['access_token']



from django.contrib import admin

from .models import PaypalPayment, PaypalTransaction, PaypalItem


class PaypalPaymentAdmin(admin.ModelAdmin):
    pass

admin.site.register(PaypalPayment, PaypalPaymentAdmin)


class PaypalTransactionAdmin(admin.ModelAdmin):
    pass

admin.site.register(PaypalTransaction, PaypalTransactionAdmin)


class PaypalItemAdmin(admin.ModelAdmin):
    pass

admin.site.register(PaypalItem, PaypalItemAdmin)
